# -*- coding: utf-8 -*-
"""
Definitions for common aid configurations

Rename to annot_cfgdef
"""
import logging
import utool as ut
import numpy as np  # NOQA


print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


# easier to type names to alias some of these options
ALIAS_KEYS = {
    # 'aids'     : 'default_aids',
    'pername': 'sample_per_name',
    'offset': 'sample_offset',
    'refrule': 'sample_rule_ref',
    'rule': 'sample_rule',
    'size': 'sample_size',
    'mingt': 'min_pername',
    'excluderef': 'exclude_reference',
}


INDEPENDENT_DEFAULTS_PARAM_INFO = [
    ut.ParamInfo('reviewed', None, valid_values=[True, False, None]),
    ut.ParamInfo(
        'minqual', None, valid_values=[None, 'junk', 'poor', 'ok', 'good', 'excellent']
    ),
    ut.ParamInfo('multiple', None, valid_values=[True, False, None]),
    ut.ParamInfo('species', None),
    ut.ParamInfo('view', None),  # TODO: allow for lists
    ut.ParamInfo('require_quality', None, valid_values=[True, False, None]),
    ut.ParamInfo('require_viewpoint', None, valid_values=[True, False, None]),
    ut.ParamInfo('is_exemplar', None, valid_values=[True, False, None]),
    ut.ParamInfo(
        'min_pername_global',
        None,
        type_=int,
        min_=0,
        help_='Keep annot if it has at least this many global names',
    ),
    ut.ParamInfo(
        'max_pername_global',
        None,
        type_=int,
        min_=0,
        help_='Keep annot if it has at most this many global names',
    ),
    ut.ParamInfo(
        'min_unixtime',
        None,
        type_=float,
        min_=0,
        help_='Remove anything before this timestamp',
    ),
    ut.ParamInfo(
        'max_unixtime',
        None,
        type_=float,
        min_=0,
        help_='Remove anything after this timestamp',
    ),
    # ut.ParamInfo('view', None),
]


INTRAGROUP_DEFAULTS_PARAM_INFO = [
    ut.ParamInfo(
        'min_pername',
        None,
        type_=int,
        min_=0,
        help_='Keeps names with at least this number of aids within the group',
    ),
    ut.ParamInfo(
        'max_pername',
        None,
        type_=int,
        min_=0,
        help_='Keeps names with at most this number of aids within the group',
    ),
]

SAMPLE_DEFAULTS_PARAM_INFO = [
    ut.ParamInfo(
        'sample_per_name', None, type_=int, min_=0, help_='Take this many annots per name'
    ),
    ut.ParamInfo(
        'sample_rule',
        'random',
        valid_values=['random', 'mintime', 'maxtime', 'qual_and_view'],
        help_='Method of samping from names',
    ),
    ut.ParamInfo(
        'sample_seed',
        0,
        type_=int,
        none_ok=True,
        help_='Random seed for sampling from names',
    ),
]

SUBINDEX_DEFAULTS_PARAM_INFO = [
    ut.ParamInfo('index', None),
]

OTHER_DEFAULTS = {
    # forces a consistnet sample size across combinations
    'force_const_size': None,
    'crossval_enc': None,
    # 'hack_extra' : None,  # hack param to make bigger db sizes
    # 'hack_imageset': None,
    # Hack out errors in test data
    # 'hackerrors'    : True,
    'hackerrors': False,
    'joinme': None,
}

# Defaults for the independent filter
# THese filters are orderless
INDEPENDENT_DEFAULTS = {
    # 'species'             : 'primary',  # specify the species
    # 'species'             : None,
    # Timedelta Params
    'require_timestamp': None,
    'require_gps': None,
    'max_timestamp': None,
    'contributor_contains': None,
    # Quality Params
    # 'require_quality'     : None,  # if True unknown qualities are removed
    # 'minqual'             : 'poor',
    'minqual': None,
    'been_adjusted': None,  # HACK PARAM
    # Viewpoint params
    # 'require_viewpoint'   : None,
    # 'view'                : None,
    'view_ext': 0,  # num viewpoints to extend in dir1 and dir2
    'view_ext1': None,  # num viewpoints to extend in dir1
    'view_ext2': None,  # num viewpoints to extend in dir2
    'is_known': None,
    # maximum number of features detected by default config
    'min_numfeat': None,
    # minimum number of features detected by default config
    'max_numfeat': None,
    'reviewed': None,
    'multiple': None,
}

# HACK
from wbia import tag_funcs  # NOQA  #

# Build Filters
filter_keys = ut.get_func_kwargs(tag_funcs.filterflags_general_tags)
for key in filter_keys:
    INDEPENDENT_DEFAULTS[key] = None

for pi in INDEPENDENT_DEFAULTS_PARAM_INFO:
    INDEPENDENT_DEFAULTS[pi.varname] = pi.default


INTRAGROUP_DEFAULTS = {
    # if True all annots must belong to the same imageset
    'same_imageset': None,
    'view_pername': None,  # formatted string filtering the viewpoints
    'min_timedelta': None,
    # minimum number of aids for each name in sample
    # 'min_pername'         : None,
    # 'max_pername'         : None,
    'min_spacedelta': None,
    'min_spacetimedelta': None,
}
for pi in INTRAGROUP_DEFAULTS_PARAM_INFO:
    INTRAGROUP_DEFAULTS[pi.varname] = pi.default

# HACK
INDEPENDENT_DEFAULTS.update(INTRAGROUP_DEFAULTS)  # hack

SUBINDEX_DEFAULTS = {
    # Final indexing
    'shuffle': False,  # randomize order before indexing
    # 'index'               : None,   # choose only a subset
}
for pi in SUBINDEX_DEFAULTS_PARAM_INFO:
    SUBINDEX_DEFAULTS[pi.varname] = pi.default

SAMPLE_DEFAULTS = {
    'sample_size': None,
    'num_names': None,
    # Gets as close to sample size without removing other props
    # Per Name / Exemplar Params
    # 'sample_per_name'     : None,  # Choos num_annots to sample from each name.
    # 'sample_rule'         : 'random',
    'sample_offset': None,  # UNUSED
    'occur_offset': None,  # UNUSED
    'name_offset': None,  # UNUSED
    'sample_occur': None,
}
for pi in SAMPLE_DEFAULTS_PARAM_INFO:
    SAMPLE_DEFAULTS[pi.varname] = pi.default

SAMPLE_REF_DEFAULTS = {
    # excludes any aids specified in a reference set (ie qaids)
    'exclude_reference': None,
    'exclude_ref_contact': None,
    'sample_rule_ref': 'random',
    # when sampling daids, choose this many correct matches per query
    'sample_per_ref_name': None,
}


# Base common settings, but some default settings will be different
# for query and database annotations
DEFAULT_AIDCFG = ut.merge_dicts(
    OTHER_DEFAULTS,
    INDEPENDENT_DEFAULTS,
    SAMPLE_DEFAULTS,
    SAMPLE_REF_DEFAULTS,
    SUBINDEX_DEFAULTS,
)
__default_aidcfg = DEFAULT_AIDCFG


def compress_aidcfg(acfg, filter_nones=False, filter_empty=False, force_noncommon=[]):
    r"""
    Idea is to add a third subconfig named `common` that is the intersection of
    `qcfg` and `dcfg`.

    Args:
        acfg (dict):

    Returns:
        dict: acfg

    CommandLine:
        # python -m wbia --tf compress_aidcfg
        python -m wbia.expt.annotation_configs --exec-compress_aidcfg --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.expt.annotation_configs import *  # NOQA
        >>> acfg = default
        >>> acfg = compress_aidcfg(acfg)
        >>> result = ('acfg = %s' % (ut.repr2(acfg),))
        >>> print(default)
        >>> print(result)
    """
    import copy

    if 'qcfg' not in acfg or 'dcfg' not in acfg:
        return acfg
    acfg = copy.deepcopy(acfg)
    common_cfg = ut.dict_intersection(acfg['qcfg'], acfg['dcfg'])
    ut.delete_keys(common_cfg, force_noncommon)
    ut.delete_keys(acfg['qcfg'], common_cfg.keys())
    ut.delete_keys(acfg['dcfg'], common_cfg.keys())
    acfg['common'] = common_cfg
    if filter_nones:
        acfg['common'] = ut.dict_filter_nones(acfg['common'])
        acfg['qcfg'] = ut.dict_filter_nones(acfg['qcfg'])
        acfg['dcfg'] = ut.dict_filter_nones(acfg['dcfg'])
    if filter_empty:
        if len(acfg['common']) == 0:
            del acfg['common']
        if len(acfg['qcfg']) == 0:
            del acfg['qcfg']
        if len(acfg['dcfg']) == 0:
            del acfg['dcfg']
    return acfg


def partition_acfg_list(acfg_list):
    for acfg in acfg_list:
        assert (
            acfg['qcfg']['_cfgname'] == acfg['dcfg']['_cfgname']
        ), 'should be the same for now'

    # Hack to make common params between q and d appear the same
    _acfg_list = [compress_aidcfg(acfg) for acfg in acfg_list]

    flat_acfg_list = flatten_acfg_list(_acfg_list)
    tup = ut.partition_varied_cfg_list(flat_acfg_list)
    flat_nonvaried_dict, flat_varied_acfg_list = tup
    nonvaried_dict = unflatten_acfgdict(flat_nonvaried_dict)
    varied_acfg_list = [unflatten_acfgdict(acfg) for acfg in flat_varied_acfg_list]
    return nonvaried_dict, varied_acfg_list


def get_varied_acfg_labels(acfg_list, mainkey='_cfgname', checkname=False):
    """
    >>> from wbia.expt.annotation_configs import *  # NOQA
    """
    # logger.info(ut.repr2(varied_acfg_list, nl=2))
    for acfg in acfg_list:
        assert acfg['qcfg'].get(mainkey, '') == acfg['dcfg'].get(
            mainkey, ''
        ), 'should be the same for now'
    cfgname_list = [acfg['qcfg'].get(mainkey, '') for acfg in acfg_list]
    if checkname and ut.allsame(cfgname_list):
        cfgname_list = [None] * len(cfgname_list)

    # Hack to make common params between q and d appear the same
    _acfg_list = [compress_aidcfg(acfg) for acfg in acfg_list]

    flat_acfg_list = flatten_acfg_list(_acfg_list)
    nonvaried_dict, varied_acfg_list = ut.partition_varied_cfg_list(flat_acfg_list)

    SUPER_HACK = True
    if SUPER_HACK:
        # SUPER HACK, recompress remake the varied list after knowing what is varied
        _varied_keys = list(
            set(
                ut.flatten(
                    [
                        list(
                            ut.flatten(
                                [list(x.keys()) for x in unflatten_acfgdict(cfg).values()]
                            )
                        )
                        for cfg in varied_acfg_list
                    ]
                )
            )
        )
        _acfg_list = [
            compress_aidcfg(acfg, force_noncommon=_varied_keys) for acfg in acfg_list
        ]
        flat_acfg_list = flatten_acfg_list(_acfg_list)
        nonvaried_dict, varied_acfg_list = ut.partition_varied_cfg_list(flat_acfg_list)

    shortened_cfg_list = [
        # {shorten_to_alias_labels(key): val for key, val in _dict.items()}
        ut.map_dict_keys(shorten_to_alias_labels, _dict)
        for _dict in varied_acfg_list
    ]
    nonlbl_keys = ut.INTERNAL_CFGKEYS
    nonlbl_keys = [prefix + key for key in nonlbl_keys for prefix in ['', 'q', 'd']]
    # hack for sorting by q/d stuff first

    def get_key_order(cfg):
        keys = [k for k in cfg.keys() if k not in nonlbl_keys]
        sortorder = [2 * k.startswith('q') + 1 * k.startswith('d') for k in keys]
        return ut.sortedby(keys, sortorder)[::-1]

    cfglbl_list = [
        ut.get_cfg_lbl(cfg, name, nonlbl_keys, key_order=get_key_order(cfg))
        for cfg, name in zip(shortened_cfg_list, cfgname_list)
    ]

    if checkname:
        cfglbl_list = [x.lstrip(':') for x in cfglbl_list]
    return cfglbl_list


def shorten_to_alias_labels(key):
    if key is None:
        return None
    search_list = list(ALIAS_KEYS.values()) + ['qcfg_', 'dcfg_', 'common_']
    repl_list = list(ALIAS_KEYS.keys()) + ['q', 'd', '']
    return ut.multi_replace(key, search_list, repl_list)


def flatten_acfg_list(acfg_list):
    """
    Returns a new config where subconfig params are prefixed by subconfig keys
    """
    flat_acfg_list = []
    for acfg in acfg_list:
        flat_dict = {
            prefix + '_' + key: val
            for prefix, subdict in acfg.items()
            for key, val in subdict.items()
        }
        flat_acfg_list.append(flat_dict)
    return flat_acfg_list


def compress_acfg_list_for_printing(acfg_list):
    r"""
    CommandLine:
        python -m wbia --tf compress_acfg_list_for_printing

    Ignore:
        >>> from wbia.expt.annotation_configs import *  # NOQA
        >>> qcfg_list = [{'f': 1, 'b': 1}, {'f': 2, 'b': 1}, {'f': 3, 'b': 1, 'z': 4}]
        >>> acfg_list = [{'qcfg': qcfg} for qcfg in qcfg_list]
        >>> nonvaried_dict, varied_dicts = compress_acfg_list_for_printing(acfg_list)
        >>> result = ('varied_dicts = %s\n' % (ut.repr2(varied_dicts),))
        >>> result += ('nonvaried_dict = %s' % (ut.repr2(nonvaried_dict),))
        >>> print(result)
    """
    flat_acfg_list = flatten_acfg_list(acfg_list)
    tup = ut.partition_varied_cfg_list(flat_acfg_list)
    nonvaried_dict, varied_acfg_list = tup
    nonvaried_compressed_dict = compress_aidcfg(
        unflatten_acfgdict(nonvaried_dict), filter_nones=True
    )
    varied_compressed_dict_list = [
        compress_aidcfg(unflatten_acfgdict(cfg), filter_empty=True)
        for cfg in varied_acfg_list
    ]
    return nonvaried_compressed_dict, varied_compressed_dict_list


def print_acfg_list(
    acfg_list,
    expanded_aids_list=None,
    ibs=None,
    combined=False,
    only_summary=False,
    **kwargs
):
    r"""
    Args:
        acfg_list (list):
        expanded_aids_list (list): (default = None)
        ibs (IBEISController):  wbia controller object(default = None)
        combined (bool): (default = False)

    CommandLine:
        python -m wbia.expt.annotation_configs --exec-print_acfg_list

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.expt.annotation_configs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> a = ['default']
        >>> acfg_list, expanded_aids_list = wbia.expt.experiment_helpers.get_annotcfg_list(
        >>>     ibs, acfg_name_list=a, verbose=0)
        >>> combined = False
        >>> result = print_acfg_list(acfg_list, expanded_aids_list, ibs, combined)
        >>> print(result)
    """
    _tup = compress_acfg_list_for_printing(acfg_list)
    nonvaried_compressed_dict, varied_compressed_dict_list = _tup

    ut.colorprint('+=== <Info acfg_list> ===', 'white')
    # logger.info('Printing acfg_list info. len(acfg_list) = %r' % (len(acfg_list),))
    logger.info('non-varied aidcfg = ' + ut.repr2(nonvaried_compressed_dict))
    seen_ = ut.ddict(list)

    # get default kwkeys for annot info
    if ibs is not None:
        annotstats_kw = kwargs.copy()
        kwkeys = ut.parse_func_kwarg_keys(ibs.get_annot_stats_dict)
        annotstats_kw.update(
            ut.argparse_dict(dict(zip(kwkeys, [None] * len(kwkeys))), only_specified=True)
        )

    hashid_list = []
    for acfgx in range(len(acfg_list)):
        acfg = acfg_list[acfgx]
        title = (
            'q_cfgname='
            + acfg['qcfg']['_cfgname']
            + ' d_cfgname='
            + acfg['dcfg']['_cfgname']
        )

        if not only_summary:
            ut.colorprint(
                '+--- acfg %d / %d -- %s ---- ' % (acfgx + 1, len(acfg_list), title),
                'gray',
            )
            logger.info('acfg = ' + ut.repr2(varied_compressed_dict_list[acfgx], si=True))

        if expanded_aids_list is not None:
            qaids, daids = expanded_aids_list[acfgx]
            key = (ut.hashstr_arr27(qaids, 'qaids'), ut.hashstr_arr27(daids, 'daids'))
            if key not in seen_:
                if ibs is not None:
                    seen_[key].append(acfgx)
                    stats_ = ibs.get_annotconfig_stats(
                        qaids, daids, verbose=False, combined=combined, **annotstats_kw
                    )
                    hashids = (
                        stats_['qaid_stats']['qhashid'],
                        stats_['daid_stats']['dhashid'],
                    )
                    hashid_list.append(hashids)
                    stats_str2 = ut.repr2(
                        stats_, si=True, nl=True, explicit=False, nobraces=False
                    )
                    if not only_summary:
                        logger.info('annot_config_stats = ' + stats_str2)
            else:
                dupindex = seen_[key]
                dupdict = varied_compressed_dict_list[dupindex[0]]
                if not only_summary:
                    logger.info('DUPLICATE of index %r' % (dupindex,))
                    logger.info('DUP OF acfg = ' + ut.repr2(dupdict, si=True))
    logger.info('hashid summary = ' + ut.repr2(hashid_list, nl=1))
    ut.colorprint('L___ </Info acfg_list> ___', 'white')


def print_acfg(acfg, expanded_aids=None, ibs=None, **kwargs):
    logger.info('acfg = ' + ut.repr2(compress_aidcfg(acfg)))
    if expanded_aids is not None:
        ibs.print_annot_stats(expanded_aids, label='expanded_aids = ', **kwargs)


def unflatten_acfgdict(flat_dict, prefix_list=['dcfg', 'qcfg']):
    acfg = {prefix: {} for prefix in prefix_list}
    for prefix in prefix_list:
        for key, val in flat_dict.items():
            if key.startswith(prefix + '_'):
                acfg[prefix][key[len(prefix) + 1 :]] = val
    return acfg


def apply_timecontrol(acfg, min_timedelta='6h', require_timestamp=True):
    return {
        'qcfg': ut.augdict(
            acfg['qcfg'],
            {'require_timestamp': require_timestamp, 'min_timedelta': min_timedelta},
        ),
        'dcfg': ut.augdict(
            acfg['dcfg'],
            {'require_timestamp': require_timestamp, 'min_timedelta': min_timedelta},
        ),
    }


def apply_qualcontrol(acfg):
    return {
        'qcfg': ut.augdict(acfg['qcfg'], {'require_quality': True}),
        'dcfg': ut.augdict(acfg['dcfg'], {'require_quality': True}),
    }


__baseline_aidcfg = ut.augdict(
    __default_aidcfg,
    {'is_known': True, 'minqual': 'ok', 'view': 'primary', 'view_ext': 1},
)


__controlled_aidcfg = ut.augdict(
    __baseline_aidcfg,
    {
        # 'require_timestamp': True,
        'view_ext': 0,
        'minqual': 'ok',
        'species': 'primary',
        'is_known': True,
    },
)

single_default = __default_aidcfg

exclude_vars = list(locals().keys())  # this line is before tests
exclude_vars.append('exclude_vars')

default = {
    'qcfg': ut.augdict(
        single_default,
        {
            # 'default_aids': (1,)
        },
    ),
    'dcfg': ut.augdict(single_default, {}),
}


default2 = {
    'qcfg': ut.augdict(default['qcfg'], {'exclude_reference': True, 'is_known': True}),
    'dcfg': ut.augdict(default['dcfg'], {'exclude_reference': True, 'is_known': True}),
}


"""
wbia -e print_acfg --db PZ_Master1 -a unctrl
"""
unctrl = uncontrolled = {
    'qcfg': ut.augdict(
        __baseline_aidcfg,
        {
            # 'default_aids': 'allgt',
            'min_pername': 2,
            'species': 'primary',
        },
    ),
    'dcfg': ut.augdict(__baseline_aidcfg, {'species': 'primary'}),
}


# Uncontrolled but comparable to controlled
unctrl_comp = {
    'qcfg': ut.augdict(
        __baseline_aidcfg,
        {
            # 'default_aids': 'allgt',
            'species': 'primary',
            'sample_per_name': 1,
            'min_pername': 2,
            'view_ext': 0,
        },
    ),
    'dcfg': ut.augdict(__baseline_aidcfg, {'species': 'primary'}),
}

"""
wbia -e print_acfg --db PZ_Master1 -a ctrl
wbia -e print_acfg --db PZ_Master1 -a unctrl ctrl::unctrl:qpername=1,qview_ext=0
wbia -e print_acfg --db PZ_Master1 -a unctrl ctrl::unctrl_comp
"""
ctrl = controlled = {
    'qcfg': ut.augdict(
        __controlled_aidcfg,
        {
            # 'default_aids': 'allgt',
            'sample_per_name': 1,
            'min_pername': 2,
        },
    ),
    'dcfg': ut.augdict(
        __controlled_aidcfg,
        {
            # 'default_aids': 'all',
            'sample_per_name': 1,
            'exclude_reference': True,
            'min_pername': 1,  # allows for singletons to be in the database
        },
    ),
}


"""
wbia -e print_acfg --db PZ_Master1 -a timectrl
"""
timectrl = timecontrolled = apply_timecontrol(ctrl)
timectrl1h = timecontrolled = apply_timecontrol(ctrl, '1h')

timectrlL = timecontrolled = apply_timecontrol(ctrl, require_timestamp=False)

"""
wbia -e print_acfg --db PZ_Master1 -a timequalctrl
"""
timequalctrl = timequalcontrolled = apply_qualcontrol(timectrl)


# Just vary the samples per name without messing with the number of annots in
# the database
varypername = {
    'qcfg': ut.augdict(
        ctrl['qcfg'],
        {
            # ensures each query will have a correct example for the groundtruth
            'min_pername': 4,
            'force_const_size': True,
        },
    ),
    'dcfg': ut.augdict(
        ctrl['qcfg'],
        {
            'sample_per_name': [1, 2, 3],
            # 'sample_per_name': [1, 3],
            # 'sample_per_ref_name': [1, 2, 3],
            # 'sample_per_ref_name': [1, 3],
            'force_const_size': True,
        },
    ),
}


varypername2 = {
    'qcfg': ut.augdict(ctrl['qcfg'], {'min_pername': 3, 'force_const_size': True}),
    'dcfg': ut.augdict(
        ctrl['dcfg'], {'sample_per_name': [1, 2], 'force_const_size': True}
    ),
}
varypername2_td = apply_timecontrol(varypername2)


"""
wbia -e print_acfg --db PZ_Master1 -a ctrl2
wbia -e print_acfg --db PZ_Master1 -a timectrl2
wbia -e rank_cmc --db PZ_Master1 -a timectrl2 -t invarbest
"""
ctrl2 = {
    'qcfg': ut.augdict(
        ctrl['qcfg'],
        {
            'min_pername': 3,
            # 'force_const_size': True,
        },
    ),
    'dcfg': ut.augdict(ctrl['dcfg'], {'sample_per_name': 2, 'force_const_size': True}),
}

timectrl2 = apply_timecontrol(ctrl2)


varypername_td = apply_timecontrol(varypername)
varypername_td1h = apply_timecontrol(varypername, '1h')
"""
wbia -e print_acfg --db PZ_Master1 -a varypername_tdqual
"""
varypername_tdqual = apply_qualcontrol(varypername_td)


"""
python -m wbia --tf get_num_annots_per_name --db PZ_Master1
wbia -e print_acfg -a varysize2 --db PZ_Master1 --verbtd --nocache
wbia -e print_acfg -a varysize2 --db NNP_MasterGIRM_core --verbtd --nocache
"""

varysize = {
    'qcfg': ut.augdict(
        __controlled_aidcfg,
        {
            # 'default_aids': 'allgt',
            'sample_size': None,
            'sample_per_name': 1,
            # 'force_const_size': True,
            'min_pername': 4,
        },
    ),
    'dcfg': ut.augdict(
        __controlled_aidcfg,
        {
            # 'default_aids': 'all',
            'sample_per_name': [1, 2, 3],
            # 'sample_per_name': [1, 3],
            'exclude_reference': True,
            'sample_size': [0.25, 0.5, 0.75],  # .95], 1.0],
            'min_pername': 1,
        },
    ),
}

"""
wbia -e print_acfg -a varysize2_td --db PZ_Master1 --verbtd --nocache
"""
varysize_td = apply_timecontrol(varysize)
varysize_td1h = apply_timecontrol(varysize, '1h')
varysize_tdqual = apply_qualcontrol(varysize_td)


varynannots = {
    'qcfg': ut.augdict(
        __controlled_aidcfg,
        {
            # 'default_aids': 'allgt',
            'sample_size': None,
            'sample_per_name': 1,
            # 'force_const_size': True,
            # 'min_pername': 4,
            'min_pername': 2,
        },
    ),
    'dcfg': ut.augdict(
        __controlled_aidcfg,
        {
            # 'default_aids': 'all',
            'sample_per_name': [1],
            # 'sample_per_name': [1, 3],
            'exclude_reference': True,
            # 'sample_size': [.01, .125, 0.25, .375, 0.5, .625, 0.75],  # , .875],  # .95], 1.0],
            # 'sample_size': [.01, .05, .125, 0.25, .375, 0.5, 0.75],  # , .875],  # .95], 1.0],
            'sample_size': [
                0.0,
                0.01,
                0.05,
                0.125,
                0.25,
                0.375,
                0.5,
                0.75,
                0.875,
                0.95,
                1.0,
            ],
            # 'sample_size': [.01, .025, .05, .125, 0.25, .375, 0.5, 0.75, .875, .95, 1.0],
            # 'sample_size': ((10 * np.logspace(0, np.log(100), num=11, base=np.e)).astype(np.int) / 1000).tolist(),
            # (10 * np.logspace(0, np.log2(100), num=11, base=2)).astype(np.int) / 1000,
            'min_pername': 1,
        },
    ),
}
varynannots_td = apply_timecontrol(varynannots)
varynannots_td1h = apply_timecontrol(varynannots, '1h')
# varysize_tdqual = apply_qualcontrol(varysize_td)


# Compare query of frontleft animals when database has only left sides
"""
wbia -e print_acfg -a viewpoint_compare --db PZ_Master1 --verbtd --nocache
python -m wbia --tf parse_acfg_combo_list -a viewpoint_compare
python -m wbia --tf get_annotcfg_list --db PZ_Master1 -a viewpoint_compare \
        --verbtd
# Check composition of names per viewpoint
python -m wbia --tf group_annots_by_multi_prop --db PZ_Master1 \
        --props=yaw_texts,name_rowids --keys1 frontleft
python -m wbia --tf get_annot_stats_dict --db PZ_Master1 \
        --per_name_vpedge=True


TODO: Need to explicitly setup the common config I think?
wbia -e print_acfg -a viewdiff:min_timedelta=1h --db PZ_Master1 --verbtd --nocache-aid
wbia --tf get_annotcfg_list -a viewdiff:min_timedelta=1h --db PZ_Master1 \
        --verbtd --nocache-aid
"""
viewpoint_compare = {
    'qcfg': ut.augdict(
        ctrl['qcfg'],
        {
            'sample_size': None,
            # To be a query you must have at least two primary1 views and at
            # least one primary view
            'view_pername': '#primary>0&#primary1>1',
            'force_const_size': True,
            'view': 'primary1',
            'sample_per_name': 1,
            # 'min_pername': 2,
        },
    ),
    'dcfg': ut.augdict(
        ctrl['dcfg'],
        {
            'view': ['primary1', 'primary'],
            'force_const_size': True,
            # To be a query you must have at least two primary1 views and at
            # least one primary view
            'view_pername': '#primary>0&#primary1>1',
            'sample_per_ref_name': 1,
            'sample_per_name': 1,
            # TODO: need to make this consistent accross both experiment modes
            'sample_size': None,
        },
    ),
}


viewdiff = vp = viewpoint_compare = {
    'qcfg': ut.augdict(
        ctrl['qcfg'],
        ut.odict(
            [
                ('sample_size', None),
                # To be a query you must have at least two primary1 views and at
                # least one primary view
                ('view_pername', '#primary>0&#primary1>0'),
                ('force_const_size', True),
                ('view', 'primary1'),
                ('sample_per_name', 1),
            ]
        ),
    ),
    'dcfg': ut.augdict(
        ctrl['dcfg'],
        {
            'view': ['primary'],
            'force_const_size': True,
            'sample_per_ref_name': 1,
            'sample_per_name': 1,  # None this seems to produce odd results
            # where the per_ref is still more then 1
            'sample_size': None,
            'view_pername': '#primary>0&#primary1>0',
        },
    ),
}


# Use tags to find a small set of difficult cases
timectrlhard = viewpoint_compare = {
    'qcfg': ut.augdict(
        timectrl['qcfg'],
        ut.odict(
            [
                ('has_any', ('needswork', 'correctable', 'mildviewpoint')),
                ('has_none', ('viewpoint', 'photobomb', 'error:viewpoint', 'quality')),
            ]
        ),
    ),
    'dcfg': ut.augdict(timectrl['dcfg'], {}),
}
"""
wbia -e print_acfg -a viewdiff --db PZ_Master1 --verbtd --nocache --per_vp=True
wbia -e print_acfg -a viewdiff_td --db PZ_Master1 --verbtd --nocache --per_vp=True
"""
viewdiff_td = apply_timecontrol(viewdiff)
viewdiff_td1h = apply_timecontrol(viewdiff, '1h')

r"""
wbia get_annotcfg_list --db Oxford -a default:qhas_any=\(query,\),dpername=2,exclude_reference=True --acfginfo --verbtd  --veryverbtd
wbia get_annotcfg_list --db Oxford -a oxford --acfginfo
('_QSUUIDS((55)qxlgljvomqpdvlny)', '_DSUUIDS((4240)vhtqsdkrwetbftis)'),

wbia draw_rank_cmc --db Oxford --save oxfordccm.png -p :proot=smk,num_words=[64000],nAssign=[1],sv_on=[False] -a oxford

"""
oxford = {
    'qcfg': ut.augdict(
        default['qcfg'],
        {
            'has_any': ('query',),
            'exclude_reference': True,
            'minqual': 'poor',
            'require_quality': False,
        },
    ),
    'dcfg': ut.augdict(
        default['dcfg'],
        {
            # 'sample_per_name': 2,
            'exclude_reference': True,
            'minqual': 'poor',
            'require_quality': False,
        },
    ),
}

# THIS IS A GOOD START
# NEED TO DO THIS CONFIG AND THEN SWITCH DCFG TO USE primary1

include_vars = list(locals().keys())  # this line is after tests

# List of all valid tests
TEST_NAMES = set(include_vars) - set(exclude_vars)



# -*- coding: utf-8 -*-
import logging
import uuid
from six.moves import range
from wbia import constants as const
from wbia.other import ibsfuncs
from wbia.control.accessor_decors import (
    adder,
    deleter,
    setter,
    getter_1to1,
    getter_1toM,
    ider,
)
import utool as ut
from wbia.control.controller_inject import make_ibs_register_decorator

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)
LBLANNOT_VALUE = 'lblannot_value'


# @ider
# def _get_all_alr_rowids(ibs):
#    all_alr_rowids = ibs.db.get_all_rowids(const.AL_RELATION_TABLE)
#    return all_alr_rowids


@register_ibs_method
@ider
def _get_all_lblannot_rowids(ibs):
    all_lblannot_rowids = ibs.db.get_all_rowids(const.LBLANNOT_TABLE)
    return all_lblannot_rowids


@register_ibs_method
@adder
def add_annot_relationship(ibs, aid_list, lblannot_rowid_list, alr_confidence_list=None):
    """
    Adds a relationship between annots and lblannots
        (annotations and labels of annotations)
    """
    if alr_confidence_list is None:
        alr_confidence_list = [0.0] * len(aid_list)
    colnames = (
        'annot_rowid',
        'lblannot_rowid',
        'alr_confidence',
    )
    params_iter = list(zip(aid_list, lblannot_rowid_list, alr_confidence_list))
    get_rowid_from_superkey = ibs.get_alrid_from_superkey
    superkey_paramx = (0, 1)  # TODO HAVE SQL GIVE YOU THESE NUMBERS
    alrid_list = ibs.db.add_cleanly(
        const.AL_RELATION_TABLE,
        colnames,
        params_iter,
        get_rowid_from_superkey,
        superkey_paramx,
    )
    return alrid_list


@register_ibs_method
@adder
def add_lblannots(
    ibs, lbltype_rowid_list, value_list, note_list=None, lblannot_uuid_list=None
):
    """Adds new lblannots (labels of annotations)
    creates a new uuid for any new pair(type, value)
    #TODO: reverse order of rowid_list value_list in input
    """
    if note_list is None:
        note_list = [''] * len(value_list)
    # Get random uuids
    if lblannot_uuid_list is None:
        lblannot_uuid_list = [uuid.uuid4() for _ in range(len(value_list))]
    colnames = ['lblannot_uuid', 'lbltype_rowid', LBLANNOT_VALUE, 'lblannot_note']
    params_iter = list(zip(lblannot_uuid_list, lbltype_rowid_list, value_list, note_list))
    get_rowid_from_superkey = ibs.get_lblannot_rowid_from_superkey
    superkey_paramx = (1, 2)
    lblannot_rowid_list = ibs.db.add_cleanly(
        const.LBLANNOT_TABLE,
        colnames,
        params_iter,
        get_rowid_from_superkey,
        superkey_paramx,
    )
    return lblannot_rowid_list


@register_ibs_method
@getter_1to1
def get_lblannot_rowid_from_superkey(ibs, lbltype_rowid_list, value_list):
    """
    Returns:
        list_ (list):  lblannot_rowid_list from the superkey (lbltype, value)
    """
    colnames = ('lblannot_rowid',)
    params_iter = zip(lbltype_rowid_list, value_list)
    where_clause = 'lbltype_rowid=? AND lblannot_value=?'
    lblannot_rowid_list = ibs.db.get_where(
        const.LBLANNOT_TABLE, colnames, params_iter, where_clause
    )
    # BIG HACK FOR ENFORCING UNKNOWN LBLANNOTS HAVE ROWID 0
    lblannot_rowid_list = [
        const.UNKNOWN_LBLANNOT_ROWID if val is None or val == const.UNKNOWN else rowid
        for rowid, val in zip(lblannot_rowid_list, value_list)
    ]
    return lblannot_rowid_list


@register_ibs_method
@deleter
def delete_annot_relations(ibs, aid_list):
    """ Deletes the relationship between an annotation and a label """
    alrids_list = ibs.get_annot_alrids(aid_list)
    alrid_list = ut.flatten(alrids_list)
    ibs.db.delete_rowids(const.AL_RELATION_TABLE, alrid_list)


@register_ibs_method
@deleter
def delete_annot_relations_oftype(ibs, aid_list, _lbltype):
    """ Deletes the relationship between an annotation and a label """
    alrids_list = ibs.get_annot_alrids_oftype(aid_list, ibs.lbltype_ids[_lbltype])
    alrid_list = ut.flatten(alrids_list)
    ibs.db.delete_rowids(const.AL_RELATION_TABLE, alrid_list)


@register_ibs_method
@deleter
def delete_lblannots(ibs, lblannot_rowid_list):
    """ deletes lblannots from the database """
    if ut.VERBOSE:
        logger.info('[ibs] deleting %d lblannots' % len(lblannot_rowid_list))
    ibs.db.delete_rowids(const.LBLANNOT_TABLE, lblannot_rowid_list)


@register_ibs_method
@getter_1to1
def get_alr_annot_rowids(ibs, alrid_list):
    """
    Args:
        alrid_list (list of rowids): annot + label relationship rows
    get the annot_rowid belonging to each relationship"""
    annot_rowids_list = ibs.db.get(const.AL_RELATION_TABLE, ('annot_rowid',), alrid_list)
    return annot_rowids_list


@register_ibs_method
def get_alr_annot_rowids_from_lblannot_rowid(ibs, lblannot_rowid_list):
    """
    This is a 1toM getter

    Get annotation rowids of labels. There may be more than one annotation
    per label.

    Args:
        lblannot_rowid_list (list): of lblannot (labels of annotations) rowids

    Returns:
        aids_list (list): of lists annotation rowids
    """
    # verbose = len(lblannot_rowid_list) > 20
    # TODO: Optimize IF POSSIBLE
    # FIXME: SLOW
    # if verbose:
    #    logger.info(ut.get_caller_name(N=list(range(0, 20))))
    where_clause = 'lblannot_rowid=?'
    params_iter = [(lblannot_rowid,) for lblannot_rowid in lblannot_rowid_list]
    aids_list = ibs.db.get_where(
        const.AL_RELATION_TABLE,
        ('annot_rowid',),
        params_iter,
        where_clause,
        unpack_scalars=False,
    )
    return aids_list


@register_ibs_method
@getter_1to1
def get_alr_confidence(ibs, alrid_list):
    """
    Args:
        alrid_list (list of rowids): annot + label relationship rows
    Returns:
        alr_confidence_list (list of rowids): confidence in an annotation relationship
    """
    alr_confidence_list = ibs.db.get(
        const.AL_RELATION_TABLE, ('alr_confidence',), alrid_list
    )
    return alr_confidence_list


@register_ibs_method
@getter_1to1
def get_alr_lblannot_rowids(ibs, alrid_list):
    """
    Args:
        alrid_list (list of rowids): annot + label relationship rows
    Returns:
        lblannot_rowids_list (list of rowids): label rowids (of annotations)
    """
    lblannot_rowids_list = ibs.db.get(
        const.AL_RELATION_TABLE, ('lblannot_rowid',), alrid_list
    )
    return lblannot_rowids_list


@register_ibs_method
@getter_1to1
def get_alrid_from_superkey(ibs, aid_list, lblannot_rowid_list):
    """
    Args:
        aid_list (list): list of annotation row-ids
        lblannot_rowid_list (list): list of lblannot row-ids
    Returns:
        alrid_list (list): annot-label relationship id list
    """
    colnames = ('annot_rowid',)
    params_iter = zip(aid_list, lblannot_rowid_list)
    where_clause = 'annot_rowid=? AND lblannot_rowid=?'
    alrid_list = ibs.db.get_where(
        const.AL_RELATION_TABLE, colnames, params_iter, where_clause
    )
    return alrid_list


@register_ibs_method
@getter_1to1
def get_annot_lblannot_value_of_lbltype(ibs, aid_list, _lbltype, lblannot_value_getter):
    """
    Returns:
        lblannot_value_list (list): a list of strings ['fred', 'sue', ...] for each chip
        identifying the animal
    """
    lbltype_dict_list = ibs.get_annot_lblannot_rowids_oftype(aid_list, _lbltype)
    DEFAULT_VALUE = const.KEY_DEFAULTS[_lbltype]
    # FIXME: Use filters and unflat maps
    lblannot_value_list = [
        lblannot_value_getter(lblannot_rowids)[0]
        if len(lblannot_rowids) > 0
        else DEFAULT_VALUE
        for lblannot_rowids in lbltype_dict_list
    ]
    return lblannot_value_list


@register_ibs_method
@getter_1to1
def get_lblannot_lbltypes_rowids(ibs, lblannot_rowid_list):
    lbltype_rowid_list = ibs.db.get(
        const.LBLANNOT_TABLE, ('lbltype_rowid',), lblannot_rowid_list
    )
    return lbltype_rowid_list


@register_ibs_method
@getter_1to1
def get_lblannot_notes(ibs, lblannot_rowid_list):
    lblannotnotes_list = ibs.db.get(
        const.LBLANNOT_TABLE, ('lblannot_note',), lblannot_rowid_list
    )
    return lblannotnotes_list


@register_ibs_method
@getter_1to1
def get_lblannot_rowid_from_uuid(ibs, lblannot_uuid_list):
    """
    UNSAFE

    Returns:
        lblannot_rowid_list from the superkey (lbltype, value)
    """
    raise AssertionError(
        'CALL TO get_lblannot_rowid_from_uuid IS UNSAFE.  USE get_lblannot_rowid_from_superkey'
    )
    colnames = ('lblannot_rowid',)
    params_iter = lblannot_uuid_list
    id_colname = 'lblannot_uuid'
    lblannot_rowid_list = ibs.db.get(
        const.LBLANNOT_TABLE, colnames, params_iter, id_colname=id_colname
    )
    return lblannot_rowid_list


@register_ibs_method
@getter_1to1
def get_lblannot_uuids(ibs, lblannot_rowid_list):
    lblannotuuid_list = ibs.db.get(
        const.LBLANNOT_TABLE, ('lblannot_uuid',), lblannot_rowid_list
    )
    return lblannotuuid_list


@register_ibs_method
@getter_1to1
def get_lblannot_values(ibs, lblannot_rowid_list, _lbltype=None):
    """
    Returns:
        text lblannots
    """
    # TODO: Remove keyword argument
    # ibsfuncs.assert_lblannot_rowids_are_type(ibs, lblannot_rowid_list,  ibs.lbltype_ids[_lbltype])
    lblannot_value_list = ibs.db.get(
        const.LBLANNOT_TABLE, (LBLANNOT_VALUE,), lblannot_rowid_list
    )
    return lblannot_value_list


@register_ibs_method
@setter
def set_alr_confidence(ibs, alrid_list, confidence_list):
    """ sets annotation-lblannot-relationship confidence """
    id_iter = ((alrid,) for alrid in alrid_list)
    val_iter = ((confidence,) for confidence in confidence_list)
    colnames = ('alr_confidence',)
    ibs.db.set(const.AL_RELATION_TABLE, colnames, val_iter, id_iter)


@register_ibs_method
@setter
def set_alr_lblannot_rowids(ibs, alrid_list, lblannot_rowid_list):
    """
    Associates whatever annotation is at row(alrid) with a new
    lblannot_rowid. (effectively changes the label value of the rowid)
    """
    id_iter = ((alrid,) for alrid in alrid_list)
    val_iter = ((lblannot_rowid,) for lblannot_rowid in lblannot_rowid_list)
    colnames = ('lblannot_rowid',)
    ibs.db.set(const.AL_RELATION_TABLE, colnames, val_iter, id_iter)


@register_ibs_method
@setter
def set_annot_lblannot_from_rowid(ibs, aid_list, lblannot_rowid_list, _lbltype):
    """ Sets items/lblannot_rowids of a list of annotations."""
    # Get the alrids_list for the aids, using the lbltype as a filter
    alrids_list = ibs.get_annot_alrids_oftype(aid_list, ibs.lbltype_ids[_lbltype])
    # Find the aids which already have relationships (of _lbltype)
    setflag_list = [len(alrids) > 0 for alrids in alrids_list]
    # Add the relationship if it doesn't exist
    aid_list_to_add = ut.get_dirty_items(aid_list, setflag_list)
    lblannot_rowid_list_to_add = ut.get_dirty_items(lblannot_rowid_list, setflag_list)
    # set the existing relationship if one already exists
    alrids_list_to_set = ut.compress(alrids_list, setflag_list)
    lblannot_rowid_list_to_set = ut.compress(lblannot_rowid_list, setflag_list)
    # Assert each annot has only one relationship of this type
    ibsfuncs.assert_singleton_relationship(ibs, alrids_list_to_set)
    alrid_list_to_set = ut.flatten(alrids_list_to_set)
    # Add the new relationships
    ibs.add_annot_relationship(aid_list_to_add, lblannot_rowid_list_to_add)
    # Set the old relationships
    ibs.set_alr_lblannot_rowids(alrid_list_to_set, lblannot_rowid_list_to_set)


@register_ibs_method
@setter
def set_annot_lblannot_from_value(ibs, aid_list, value_list, _lbltype, ensure=True):
    """
    Associates the annot and lblannot of a specific type and value
    Adds the lblannot if it doesnt exist.
    Wrapper around convenience function for set_annot_from_lblannot_rowid
    """
    assert value_list is not None
    assert _lbltype is not None
    if ensure:
        pass
    # a value consisting of an empty string or all spaces is set to the default
    DEFAULT_VALUE = const.KEY_DEFAULTS[_lbltype]
    EMPTY_KEY = const.EMPTY_KEY
    # setting a name to DEFAULT_VALUE or EMPTY is equivalent to unnaming it
    value_list_ = [
        DEFAULT_VALUE if value.strip() == EMPTY_KEY else value for value in value_list
    ]
    notdefault_list = [value != DEFAULT_VALUE for value in value_list_]
    aid_list_to_delete = ut.get_dirty_items(aid_list, notdefault_list)
    # Set all the valid valids
    aids_to_set = ut.compress(aid_list, notdefault_list)
    values_to_set = ut.compress(value_list_, notdefault_list)
    ibs.delete_annot_relations_oftype(aid_list_to_delete, _lbltype)
    # remove the relationships that have now been unnamed
    # Convert names into lblannot_rowid
    # FIXME: This function should not be able to set label realationships
    # to labels that have not been added!!
    # This is an inefficient way of getting lblannot_rowids!
    lbltype_rowid_list = [ibs.lbltype_ids[_lbltype]] * len(values_to_set)
    # auto ensure
    lblannot_rowid_list = ibs.add_lblannots(lbltype_rowid_list, values_to_set)
    # Call set_annot_from_lblannot_rowid to finish the conditional adding
    ibs.set_annot_lblannot_from_rowid(aids_to_set, lblannot_rowid_list, _lbltype)


@register_ibs_method
def set_lblannot_notes(ibs, lblannot_rowid_list, value_list):
    """
    Updates the value for lblannots. Note this change applies to
    all annotations related to this lblannot_rowid
    """
    id_iter = ((rowid,) for rowid in lblannot_rowid_list)
    val_list = ((value,) for value in value_list)
    ibs.db.set(const.LBLANNOT_TABLE, ('lblannot_note',), val_list, id_iter)


@register_ibs_method
def set_lblannot_values(ibs, lblannot_rowid_list, value_list):
    """
    Updates the value for lblannots. Note this change applies to
    all annotations related to this lblannot_rowid
    """
    id_iter = ((rowid,) for rowid in lblannot_rowid_list)
    val_list = ((value,) for value in value_list)
    ibs.db.set(const.LBLANNOT_TABLE, (LBLANNOT_VALUE,), val_list, id_iter)


@register_ibs_method
@getter_1toM
def get_annot_alrids(ibs, aid_list):
    """FIXME: __name__
    Get all the relationship ids belonging to the input annotations
    if lblannot lbltype is specified the relationship ids are filtered to
    be only of a specific lbltype/category/type
    """
    alrids_list = ibs.db.get(
        const.AL_RELATION_TABLE,
        ('alr_rowid',),
        aid_list,
        id_colname='annot_rowid',
        unpack_scalars=False,
    )
    return alrids_list


@register_ibs_method
@getter_1toM
def get_annot_alrids_oftype(ibs, aid_list, lbltype_rowid):
    """
    Get all the relationship ids belonging to the input annotations where the
    relationship ids are filtered to be only of a specific lbltype/category/type
    """
    alrids_list = ibs.get_annot_alrids(aid_list)
    # Get lblannot_rowid of each relationship
    lblannot_rowids_list = ibsfuncs.unflat_map(ibs.get_alr_lblannot_rowids, alrids_list)
    # Get the type of each lblannot
    lbltype_rowids_list = ibsfuncs.unflat_map(
        ibs.get_lblannot_lbltypes_rowids, lblannot_rowids_list
    )
    # only want the nids of individuals, not species, for example
    valids_list = [
        [typeid == lbltype_rowid for typeid in rowids] for rowids in lbltype_rowids_list
    ]
    alrids_list = [
        ut.compress(alrids, valids) for alrids, valids in zip(alrids_list, valids_list)
    ]

    def resolution_func_first(alrid_list):
        return [alrid_list[0]]

    alrids_list = [
        resolution_func_first(alrid_list) if len(alrid_list) > 1 else alrid_list
        for alrid_list in alrids_list
    ]
    assert all([len(alrid_list) < 2 for alrid_list in alrids_list]), (
        'More than one type per lbltype.  ALRIDS: '
        + str(alrids_list)
        + ', ROW: '
        + str(lbltype_rowid)
        + ', KEYS:'
        + str(ibs.lbltype_ids)
    )
    return alrids_list


@register_ibs_method
@getter_1toM
def get_annot_lblannot_rowids(ibs, aid_list):
    """
    Returns:
        list_ (list): the name id of each annotation."""
    # Get all the annotation lblannot relationships
    # filter out only the ones which specify names
    alrids_list = ibs.get_annot_alrids(aid_list)
    lblannot_rowids_list = ibsfuncs.unflat_map(ibs.get_alr_lblannot_rowids, alrids_list)
    return lblannot_rowids_list


@register_ibs_method
@getter_1toM
def get_annot_lblannot_rowids_oftype(ibs, aid_list, _lbltype=None):
    """
    Returns:
        list_ (list): the name id of each annotation."""
    # Get all the annotation lblannot relationships
    # filter out only the ones which specify names
    assert _lbltype is not None, 'should be using lbltype_rowids anyway'
    alrids_list = ibs.get_annot_alrids_oftype(aid_list, ibs.lbltype_ids[_lbltype])
    lblannot_rowids_list = ibsfuncs.unflat_map(ibs.get_alr_lblannot_rowids, alrids_list)
    return lblannot_rowids_list



# -*- coding: utf-8 -*-
#!/usr/bin/env python  # NOQA
"""Converts a GGR-style raw data to IBEIS database."""
import logging
from wbia.detecttools.directory import Directory
from os.path import join, exists
import utool as ut
import wbia

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def _fix_ggr2018_directory_structure(ggr_path):

    # Manual fixes for bad directories

    src_uri = join(ggr_path, 'Clarine\\ Plane\\ Kurungu/')
    dst_uri = join(ggr_path, '231/')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '231B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(
        ggr_path,
        'Alex\\ Peltier\\ -\\ Plane\\ -\\ Ngurnit/giraffe\\ grevy\\ count\\ feb\\ 18/',
    )
    dst_uri = join(ggr_path, '232/')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '232B/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-2])
    ut.delete(src_uri)

    src_uri = join(
        ggr_path, 'Mint\\ Media\\ Footage', 'Mpala\\ day\\ 1\\ spark', 'PANORAMA/'
    )
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, 'Mint\\ Media\\ Footage', 'Mpala\\ day\\ 1/')
    dst_uri = join(ggr_path, '233/')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '233B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, 'Mint\\ Media\\ Footage', 'Mpala\\ day\\ 1\\ spark/')
    dst_uri = join(ggr_path, '233', '233B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, 'Mint\\ Media\\ Footage', 'Mpala\\ day2\\ /')
    dst_uri = join(ggr_path, '233', '233B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, 'Mint\\ Media\\ Footage', 'Mpala\\ day\\ 2\\ spark/')
    dst_uri = join(ggr_path, '233', '233B/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-2])
    ut.delete(src_uri)

    src_uri = join(ggr_path, '103\\ \\(1\\)/')
    dst_uri = join(ggr_path, '103/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '103\\ \\(ccef473b\\)/')
    dst_uri = join(ggr_path, '103/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '108\\ \\(1\\)/')
    dst_uri = join(ggr_path, '108/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '226A\\ \\(Shaba\\ Funan\\ Camp\\)/')
    dst_uri = join(ggr_path, '226/')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '226A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '121/*.*')
    dst_uri = join(ggr_path, '121', '121A/')
    ut.rsync(src_uri, dst_uri)
    for src_filepath in ut.glob(src_uri.replace('\\', '')):
        ut.delete(src_filepath)

    src_uri = join(ggr_path, '54', '54A\\(16\\)/')
    dst_uri = join(ggr_path, '54', '54A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '54', '54B\\(16\\)/')
    dst_uri = join(ggr_path, '54', '54B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '87', '87/')
    dst_uri = join(ggr_path, '87', '87A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '223', 'A/')
    dst_uri = join(ggr_path, '223', '223A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '223', 'B/')
    dst_uri = join(ggr_path, '223', '223B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '14', '15A/')
    dst_uri = join(ggr_path, '14', '14A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '73/')
    dst_uri = join(ggr_path, '85/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '117', '115A/')
    dst_uri = join(ggr_path, '117', '117A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '200\\ A/')
    dst_uri = join(ggr_path, '200', '200A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '200\\ B/')
    dst_uri = join(ggr_path, '200', '200B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '200\\ F/')
    dst_uri = join(ggr_path, '200', '200F/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '200A/')
    dst_uri = join(ggr_path, '201/')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '201A/')
    ut.rsync(src_uri, dst_uri)
    # ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '201\\ E/')
    dst_uri = join(ggr_path, '201', '201E/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '201\\ F/')
    dst_uri = join(ggr_path, '201', '201F/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '200A/')
    dst_uri = join(ggr_path, '202/')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '202A/')
    ut.rsync(src_uri, dst_uri)
    # ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '202\\ B/')
    dst_uri = join(ggr_path, '202', '202B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '200', '202\\ F/')
    dst_uri = join(ggr_path, '202', '202F/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '230', '230A', 'El\\ Karama/*.*')
    dst_uri = join(ggr_path, '230', '230A/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-1])
    ut.delete(src_uri)

    src_uri = join(ggr_path, '136', '136B', '136B\\ Grevys\\ Rally/*.*')
    dst_uri = join(ggr_path, '136', '136B/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-1])
    ut.delete(src_uri)

    src_uri = join(ggr_path, '160', '160E', '104DUSIT')
    if exists(src_uri):
        direct = Directory(src_uri, recursive=False)
        filename_list = direct.files()
        for filename in sorted(filename_list):
            dst_uri = filename.replace('104DUSIT/', '').replace('.JPG', '_.JPG')
            assert not exists(dst_uri)
            ut.rsync(filename, dst_uri)
        ut.delete(src_uri)

    src_uri = join(ggr_path, '222', '222B', '102DUSIT')
    if exists(src_uri):
        direct = Directory(src_uri, recursive=False)
        filename_list = direct.files()
        for filename in sorted(filename_list):
            dst_uri = filename.replace('102DUSIT/', '').replace('.JPG', '_.JPG')
            assert not exists(dst_uri)
            ut.rsync(filename, dst_uri)
        ut.delete(src_uri)

    # Errors found by QR codes

    # No conflicts
    src_uri = join(ggr_path, '5', '5A/')
    dst_uri = join(ggr_path, '5', '5B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '14', '14A/')
    dst_uri = join(ggr_path, '14', '14B/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '118', '118A/')
    dst_uri = join(ggr_path, '192')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '192A/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-2])
    ut.delete(src_uri)

    src_uri = join(ggr_path, '119', '119A/')
    dst_uri = join(ggr_path, '189')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '189A/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-2])
    ut.delete(src_uri)

    src_uri = join(ggr_path, '120', '120A/')
    dst_uri = join(ggr_path, '190')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '190A/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-2])
    ut.delete(src_uri)

    src_uri = join(ggr_path, '138', '138C/')
    dst_uri = join(ggr_path, '169', '169C/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri)

    # Conflicts - Move first

    src_uri = join(ggr_path, '115', '115A/')
    dst_uri = join(ggr_path, '191')
    ut.ensuredir(dst_uri)
    dst_uri = join(dst_uri, '191A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    src_uri = join(ggr_path, '148', '148A/')
    dst_uri = join(ggr_path, '149', '149A-temp/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    # Conflicts - Move second

    src_uri = join(ggr_path, '117', '117A/')
    dst_uri = join(ggr_path, '115', '115A/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-2])
    ut.delete(src_uri)

    src_uri = join(ggr_path, '149', '149A/')
    dst_uri = join(ggr_path, '148', '148A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    # Conflicts - Move third

    src_uri = join(ggr_path, '149', '149A-temp/')
    dst_uri = join(ggr_path, '149', '149A/')
    ut.rsync(src_uri, dst_uri)
    ut.delete(src_uri.replace('\\', ''))

    # Conflicts - Merge third

    src_uri = join(ggr_path, '57', '57A/')
    dst_uri = join(ggr_path, '25', '25A/')
    ut.rsync(src_uri, dst_uri)
    src_uri = src_uri.replace('\\', '')
    src_uri = '/'.join(src_uri.split('/')[:-2])
    ut.delete(src_uri)


def convert_ggr2018_to_wbia(
    ggr_path, dbdir=None, purge=True, dry_run=False, apply_updates=True, **kwargs
):
    r"""Convert the raw GGR2 (2018) data to an wbia database.

    Args
        ggr_path (str): Directory to folder *containing* raw GGR 2018 data
        dbdir (str): Output directory

    CommandLine:
        python -m wbia convert_ggr2018_to_wbia

    Example:
        >>> # SCRIPT
        >>> from wbia.dbio.ingest_ggr import *  # NOQA
        >>> default_ggr_path = join('/', 'data', 'wbia', 'GGR2', 'GGR2018data')
        >>> default_dbdir = join('/', 'data', 'wbia', 'GGR2-IBEIS')
        >>> dbdir = ut.get_argval('--dbdir', type_=str, default=default_dbdir)
        >>> ggr_path = ut.get_argval('--ggr', type_=str, default=default_ggr_path)
        >>> result = convert_ggr2018_to_wbia(ggr_path, dbdir=dbdir, purge=False, dry_run=True, apply_updates=False)
        >>> print(result)
    """
    ALLOWED_NUMBERS = list(range(1, 250))
    ALLOWED_LETTERS = ['A', 'B', 'C', 'D', 'E', 'F']

    ################################################################################

    if apply_updates:
        _fix_ggr2018_directory_structure(ggr_path)

    ################################################################################

    blacklist_filepath_set = set(
        [
            join(ggr_path, 'Cameras info.numbers'),
            join(ggr_path, 'Cameras info.xlsx'),
            join(ggr_path, 'GGR_photos_MRC_29.1.18.ods'),
            join(ggr_path, 'Cameras info-2.numbers'),
        ]
    )

    # Check root files
    direct = Directory(ggr_path)
    for filepath in direct.files(recursive=False):
        try:
            assert filepath in blacklist_filepath_set
            ut.delete(filepath)
        except AssertionError:
            logger.info('Unresolved root file found in %r' % (filepath,))
            continue

    ################################################################################

    if purge:
        ut.delete(dbdir)
    ibs = wbia.opendb(dbdir=dbdir)

    ################################################################################

    # Check folder structure
    assert exists(ggr_path)
    direct = Directory(ggr_path, recursive=0)
    direct1_list = direct.directories()
    direct1_list.sort(key=lambda x: int(x.base()), reverse=False)
    for direct1 in direct1_list:
        if not dry_run:
            logger.info('Processing directory: %r' % (direct1,))
        base1 = direct1.base()

        try:
            int(base1)
        except ValueError:
            logger.info('Error found in %r' % (direct1,))
            continue

        try:
            assert len(direct1.files(recursive=False)) == 0
        except AssertionError:
            logger.info('Files found in %r' % (direct1,))
            continue

        seen_letter_list = []
        direct1_ = Directory(direct1.absolute_directory_path, recursive=0)
        direct2_list = direct1_.directories()
        direct2_list.sort(key=lambda x: x.base(), reverse=False)
        for direct2 in direct2_list:
            base2 = direct2.base()

            try:
                assert base2.startswith(base1)
            except AssertionError:
                logger.info(
                    'Folder name heredity conflict %r with %r' % (direct2, direct1)
                )
                continue

            try:
                assert len(base2) >= 2
                assert ' ' not in base2
                number = base2[:-1]
                letter = base2[-1]
                number = int(number)
                letter = letter.upper()
                assert number in ALLOWED_NUMBERS
                assert letter in ALLOWED_LETTERS
                seen_letter_list.append(letter)
            except ValueError:
                logger.info('Error found in %r' % (direct2,))
                continue
            except AssertionError:
                logger.info('Folder name format error found in %r' % (direct2,))
                continue

            direct2_ = Directory(
                direct2.absolute_directory_path, recursive=True, images=True
            )
            try:
                assert len(direct2_.directories()) == 0
            except AssertionError:
                logger.info('Folders exist in file only level %r' % (direct2,))
                continue

            filepath_list = sorted(direct2_.files())

            if not dry_run:
                try:
                    gid_list = ibs.add_images(filepath_list)
                    gid_list = ut.filter_Nones(gid_list)
                    gid_list = sorted(list(set(gid_list)))

                    imageset_text = 'GGR2,%d,%s' % (number, letter)
                    note_list = [
                        '%s,%05d' % (imageset_text, index + 1)
                        for index, gid in enumerate(gid_list)
                    ]
                    ibs.set_image_notes(gid_list, note_list)
                    ibs.set_image_imagesettext(gid_list, [imageset_text] * len(gid_list))
                except Exception as ex:  # NOQA
                    ut.embed()

        seen_letter_set = set(seen_letter_list)
        try:
            assert len(seen_letter_set) == len(seen_letter_list)
        except AssertionError:
            logger.info(
                'Duplicate letters in %r with letters %r' % (direct1, seen_letter_list)
            )
            continue

        try:
            assert 'A' in seen_letter_set
        except AssertionError:
            logger.info('WARNING: A camera not found in %r' % (direct1,))
            continue

    return ibs
